
/*
	Tims_Electronic_Point_Mapper.ino
	By Tim Jackson.1960

	Creadits:
		Arduino.
		LiquidCrystal_I2C based on work by DFRobot.

	This is code for: Tim's Electronic Point Mapper.
	The code is written for ATMEGA based boards, where the double is only 4 bytes.
	I am using 3D Printed Parts with three 49E Linear Hall Effect Sensor and two Magnets on each sensor.
	I am using an Arduino NANO to calculate the postion of the sylus from the values recived from the two 49E Linear Hall Effect Sensors.
	The sensor is linear: 3 mV/GS, but to get the angle a Sine of the value is calculated.

  Note!
  I am using pin A6 and A7 on the Arduino NANO.
  If using other boards, the Pin Assignment may need changing.

	S=O/H C=A/H T=O/A
	Degrees to Radians = degrees * (PI / 180)
	Radians to Degrees = radians * 180.0 / pi

*/

#include <Wire.h> 
#include <LiquidCrystal_I2C.h>

//#define DEBUG					//	Activates Serial Printing. (Comment out to turn off DBUG)
#define ARM_LENGTH		200		//	Length between centres of Arms.
/*
	Pins
*/
#define PEN_BUTTON		16		//	Pin A2	Button.
#define Hall_49R_Pin_03	17		//	Pin A3	define Hall-Effect sensor Pin.
//	18	A4 = SDA
//	19	A5 = SCL
#define Hall_49R_Pin_02	20		//	Pin A6	define Hall-Effect sensor Pin.
#define Hall_49R_Pin_01	21		//	Pin A7	define Hall-Effect sensor Pin.
/*
	Button
		Change Pen State.
*/
volatile bool Button = false;
/*
	Sensors

	Note!
		During Calibration, if Values are the wrong way around (small / large), then magnets are the wrong way around.
*/
#define CAL_HALL_0D_01			215.15											//	Value created by NANO.
#define CAL_HALL_45D_01			310												//	Small Value of hall-effect sensor 1 at 45 angle. Move 1st Arm.
#define CAL_HALL_135D_01		768												//	Large Value of hall-effect sensor 1 at 135 angle. Move 1st Arm Down.
#define CAL_HALL_180D_01		862.85											//	Value created by NANO.
#define CAL_HALL_RANGE_01		(CAL_HALL_180D_01 - CAL_HALL_0D_01)				//	Highest CAL - Lowest CAL.
#define CAL_HALL_RAD_01			((double)CAL_HALL_RANGE_01 / 2)					//	Used in the sine H value to get angle.
#define ANGLE_BIOS_01			0												//	Angle correction at calibration point, after by NANO correction.
#define ARM_ANGLE_OFFSET_01		-90												//	Off set alngle of first arm from vertical CW.
unsigned int Sensor_Value_01 = CAL_HALL_0D_01;									//	Variable to hold Sensor value.
double Hall_Angle_01 = 0;														//	Variable to hold Hall 1 Angle value.
double Hall_R_01 = 0;															//	Variable to hold Hall 1 Radius value.
double Hall_Z_01 = 0;															//	Variable to hold Hall 1 Z value.
double AdjustedAngle_01 = Hall_Angle_01 + ANGLE_BIOS_01 + ARM_ANGLE_OFFSET_01;	//	Used for Trigonometry.

#define CAL_HALL_0D_02			207.11											//	Value created by NANO.
#define CAL_HALL_45D_02			303												//	Small Value of hall-effect sensor 2 at 45 angle. Move away from 1st Arm.
#define CAL_HALL_135D_02		766												//	Large Value of hall-effect sensor 2 at 135 angle. Move close to 1st Arm.
#define CAL_HALL_180D_02		861.89											//	Value created by NANO.
#define CAL_HALL_RANGE_02		(CAL_HALL_180D_02 - CAL_HALL_0D_02)				//	Highest CAL - Lowest CAL.
#define CAL_HALL_RAD_02			((double)CAL_HALL_RANGE_02 / 2)					//	Used in the sine H value to get angle.
#define ANGLE_BIOS_02			0												//	Angle correction at calibration point, after by NANO correction.
#define ARM_ANGLE_OFFSET_02		-30												//	Off set alngle of swcond arm from vertical CW.
#define STYLUS_R_OFFSET	-50														//	Off set of the radius to Stylus.
#define STYLUS_Z_OFFSET	-40														//	Off set of the radius to Stylus.
unsigned int Sensor_Value_02 = CAL_HALL_0D_02;									//	Variable to hold Sensor value.
double Hall_Angle_02 = 0;														//	Variable to hold Hall 2 Angle value.
double Hall_R_02 = 0;															//	Variable to hold Hall 2 Radius value.
double Hall_Z_02 = 0;															//	Variable to hold Hall 2 Z value.
double AdjustedAngle_02 = Hall_Angle_02 + ANGLE_BIOS_02 + ARM_ANGLE_OFFSET_02;	//	Used for Trigonometry.

#define CAL_HALL_0D_03			148.19											//	Value created by NANO.
#define CAL_HALL_45D_03			265												//	Small Value of hall-effect sensor 3 at +45 angle. Move all the way tp the right.
#define CAL_HALL_135D_03		829												//	Large Value of hall-effect sensor 3 at -45 angle. Move all the way tp the left.
#define CAL_HALL_180D_03		945.81											//	Value created by NANO.
#define CAL_HALL_RANGE_03		(CAL_HALL_180D_03 - CAL_HALL_0D_03)				//	Highest CAL - Lowest CAL.
#define CAL_HALL_RAD_03			((double)CAL_HALL_RANGE_03 / 2)					//	Used in the sine H value to get angle.
#define ANGLE_BIOS_03			0												//	Angle correction at calibration point, after by NANO correction.
unsigned int Sensor_Value_03 = CAL_HALL_0D_03;									//	Variable to hold Sensor value.
double Hall_Angle_03 = 0;														//	Variable to hold Hall 3 Angle value.

/*
	Output
*/
double Sylus_R = 0;						//	Variable to hold Sylus R Coordinate.
double Sylus_A = 0;						//	Variable to hold Sylus A Coordinate.
double Sylus_Z = 0;						//	Variable to hold Sylus Z Coordinate.
bool Pen_Down = true;					//	State of Pen.

LiquidCrystal_I2C lcd(0x27, 16, 2);		//	(0x20 to 0x27)	Set to 0x27, 16 chars and 2 line display.
/*
	Timing
*/
unsigned long NowTime = millis();
unsigned long IntevalTime = 200;
unsigned long NextTime = NowTime + IntevalTime;

void  setup()
{

	Serial.begin(115200);					//	Start Serial.

	pinMode(Hall_49R_Pin_01, INPUT);		//	Define as Input.
	pinMode(Hall_49R_Pin_02, INPUT);		//	Define as Input.
	pinMode(Hall_49R_Pin_03, INPUT);		//	Define as Input.
	pinMode(PEN_BUTTON, INPUT);			//  Define as Input.

	//00000|PCIE2|PCIE1|PCIE0
	PCICR = 0b00000010; // Pin Change Interrupt Enable.
	//PCMSK0=PCINT 7,6,5,4,3,2,1,0.(Pins= crystal, crystal, 13, 12, 11, 10, 9, 8)
	//PCMSK1=PCINT 14,13,12,11,10,9,8.(Pins= reset, A5,A4, A3, A2, A1, A0)
	//PCMSK2=PCINT 23,22,21,20,19,18,17,16.(Pins= 7, 6, 5, 4, 3, 2, 1, 0)
	PCMSK1 = 0b00000100; // Enable Pin Change Interrupt for A2.
	sei();//enable interupts

	lcd.init();								//	Start LCD.
	lcd.backlight();						//	Turn on back light.
	lcd.setCursor(0, 0);					//	Set Cursor at the begining of line 0 (Top Line).
	lcd.print("R");							//	Display Label for R on top line.
	lcd.setCursor(9, 0);					//	Set Cursor to the right a little of line 0 (Top Line).
	lcd.print("A");							//	Display Label for A on top line.
	lcd.setCursor(0, 1);					//	Set Cursor at the begining of line 1 (Bottom Line).
	lcd.print("Z");							//	Display Label for Z on bottom line.

#ifdef DEBUG
	lcd.setCursor(11, 1);					//	Set Cursor near the end of line 0 (top Line).
	lcd.print("DEBUG");						//	Display DEBUG on top line.
#else
	Do_Button();							//	Display Pen State.	
#endif // DEBUG

}
void  loop() {

	NowTime = millis();

	if (Button) { Do_Button(); }

	Read_Sensor_Values();
	Calc_Angle_Hall_01();								//	Do sub routeen for angle first arm.
	Calc_Angle_Hall_02();								//	Do sub routeen for angle second arm.
	Calc_Angle_Hall_03();								//	Do sub routeen for angle rotation pivot.
	Calc_RZ_Hall_01();									//	Do sub routeen for first Cooardinates.
	Calc_RZ_Hall_02();									//	Do sub routeen for second Cooardinates.
	Calculate_Coordinates();

#ifdef DEBUG
	Calc_0_180();
#endif // DEBUG

	/*
		Send data at intervals, don't want to clog the Serial.
		Don't want it send this data when in DEBUG mode.
	*/
	if (NowTime > NextTime) {
		NextTime = NowTime + IntevalTime;
		Send_Data();
	}

}
ISR(PCINT1_vect) {

	// If interrupt is triggered by the Pen Button
	if (!digitalRead(PEN_BUTTON)) { Button = true; }

}
/*
	Calculate 0 and 180 from Sector Quadrants

	H=O/S
	Degrees to Radians = degrees * (PI / 180)

	Segment_Chord = CAL_HALL_135D - CAL_HALL_45D
	Opp = Segment_Chord / 2
	Rad = Opp / Sin(45 * (PI / 180))
	Adjustment = Opp - Rad

	0_Deg = CAL_HALL_45D - Adjustment
	180_Deg = CAL_HALL_135D + Adjustment


*/
void Calc_0_180() {

	double Opp = (double)(CAL_HALL_135D_01 - CAL_HALL_45D_01) / 2.0;
	double Rad = (double)(Opp / sin(45 * (PI / 180)));
	double Adjustment = (double)(Rad - Opp);

	Serial.println();									//	Send a new line to serial to seperate values.
	Serial.print("Sensor 1 Calculated 0 value: ");		//	Send value to serial.
	Serial.println(CAL_HALL_45D_01 - Adjustment);		//	Send value to serial.
	Serial.print("Sensor 1 Calculated 180 value: ");	//	Send value to serial.
	Serial.println(CAL_HALL_135D_01 + Adjustment);		//	Send value to serial.

	Opp = (double)(CAL_HALL_135D_02 - CAL_HALL_45D_02) / 2.0;
	Rad = (double)(Opp / sin(45 * (PI / 180)));
	Adjustment = (double)(Rad - Opp);

	Serial.println();									//	Send a new line to serial to seperate values.
	Serial.print("Sensor 2 Calculated 0 value: ");		//	Send value to serial.
	Serial.println(CAL_HALL_45D_02 - Adjustment);		//	Send value to serial.
	Serial.print("Sensor 2 Calculated 180 value: ");	//	Send value to serial.
	Serial.println(CAL_HALL_135D_02 + Adjustment);		//	Send value to serial.

	Opp = (double)(CAL_HALL_135D_03 - CAL_HALL_45D_03) / 2.0;
	Rad = (double)(Opp / sin(45 * (PI / 180)));
	Adjustment = (double)(Rad - Opp);

	Serial.println();									//	Send a new line to serial to seperate values.
	Serial.print("Sensor 3 Calculated 0 value: ");		//	Send value to serial.
	Serial.println(CAL_HALL_45D_03 - Adjustment);		//	Send value to serial.
	Serial.print("Sensor 3 Calculated 180 value: ");	//	Send value to serial.
	Serial.println(CAL_HALL_135D_03 + Adjustment);		//	Send value to serial.

	Serial.println();									//	Send a new line to serial to seperate values.

}
/*
	Read Sensor Values

*/
void Read_Sensor_Values() {

	Sensor_Value_01 = 0;
	Sensor_Value_02 = 0;
	Sensor_Value_03 = 0;

	for (size_t i = 0; i < 10; i++)
	{
		Sensor_Value_01 += analogRead(Hall_49R_Pin_01);		//	Read the value from sensor.
		Sensor_Value_02 += analogRead(Hall_49R_Pin_02);		//	Read the value from sensor.
		Sensor_Value_03 += analogRead(Hall_49R_Pin_03);		//	Read the value from sensor.
	}

	Sensor_Value_01 = Sensor_Value_01 / 10;
	Sensor_Value_02 = Sensor_Value_02 / 10;
	Sensor_Value_03 = Sensor_Value_03 / 10;

#ifdef DEBUG
	Serial.println();									//	Send a new line to serial to seperate values.
	Serial.println(Sensor_Value_01);					//	Send value to serial.
	Serial.println(Sensor_Value_02);					//	Send value to serial.
	Serial.println(Sensor_Value_03);					//	Send value to serial.
	delay(1000);										//	Wait a little for things to happen and be able to read results.
#endif // DEBUG

}
/*
	Calculate the Hall Angle using Value from Sensor.

		S=O/H
		Radians to Degrees = radians * 180.0 / pi

		O = SensorValue - CAL_HALL_0D - CAL_HALL_RAD
		H = CAL_HALL_RAD
		Hall_Angle in radians = asin(O / H)
		Hall_Angle in degrees = Hall_Angle in radians * 180.0 / pi;

*/
void Calc_Angle_Hall_01() {

	double O = (double)Sensor_Value_01 - CAL_HALL_0D_01 - CAL_HALL_RAD_01;
	double H = (double)O / CAL_HALL_RAD_01;
	Hall_Angle_01 = 90.0 + (asin(H) * 180.0 / PI);

#ifdef DEBUG
	Serial.print("Hall_Angle_01 ");
	Serial.println(Hall_Angle_01 + ANGLE_BIOS_01, 4);
#endif // DEBUG

}
void Calc_Angle_Hall_02() {

	double O = (double)Sensor_Value_02 - CAL_HALL_0D_02 - CAL_HALL_RAD_02;
	double H = (double)O / CAL_HALL_RAD_02;
	Hall_Angle_02 = 90.0 + (asin(H) * 180.0 / PI);

#ifdef DEBUG
	Serial.print("Hall_Angle_02 ");
	Serial.println(Hall_Angle_02 + ANGLE_BIOS_02, 4);
#endif // DEBUG

}
void Calc_Angle_Hall_03() {

	double O = (double)Sensor_Value_03 - CAL_HALL_0D_03 - CAL_HALL_RAD_03;
	double H = (double)O / CAL_HALL_RAD_03;
	Hall_Angle_03 = 90.0 + (asin(H) * 180.0 / PI);

#ifdef DEBUG
	Serial.print("Hall_Angle_03 ");
	Serial.println(Hall_Angle_03 + ANGLE_BIOS_03, 4);
#endif // DEBUG

}
/*
	Calculate the RZ (Radius Z) Position of the end of First Arm.
	Note!
		Angles have off-set

	S=O/H C=A/H T=O/A
	Degrees to Radians = degrees * (PI / 180)

	S=O/H
	O = S*H
	Z = sin((Hall_Angle_01 - 90) * (PI / 180)) * ARM_LENGTH

	C=A/H
	A = C*H
	R = cos((Hall_Angle_01 - 90) * (PI / 180)) * ARM_LENGTH





 */
void Calc_RZ_Hall_01() {

	AdjustedAngle_01 = Hall_Angle_01 + ANGLE_BIOS_01 + ARM_ANGLE_OFFSET_01;

	Hall_R_01 = cos(AdjustedAngle_01 * (PI / 180)) * ARM_LENGTH;
	Hall_Z_01 = -1 * (sin(AdjustedAngle_01 * (PI / 180)) * ARM_LENGTH);

#ifdef DEBUG
	Serial.print("AdjustedAngle_01 ");
	Serial.println(AdjustedAngle_01, 4);
	Serial.print("Hall_R_01 ");
	Serial.println(Hall_R_01, 4);
	Serial.print("Hall_Z_01 ");
	Serial.println(Hall_Z_01, 4);
#endif // DEBUG

}
void Calc_RZ_Hall_02() {

	AdjustedAngle_02 = 180 - Hall_Angle_02 + ANGLE_BIOS_02 + ARM_ANGLE_OFFSET_02 - AdjustedAngle_01;

	Hall_R_02 = -1 * (cos(AdjustedAngle_02 * (PI / 180)) * ARM_LENGTH);
	Hall_Z_02 = -1 * (sin(AdjustedAngle_02 * (PI / 180)) * ARM_LENGTH);

	Hall_R_02 += STYLUS_R_OFFSET;
	Hall_Z_02 += STYLUS_Z_OFFSET;

#ifdef DEBUG
	Serial.print("AdjustedAngle_02 ");
	Serial.println(AdjustedAngle_02, 4);
	Serial.print("Hall_R_02 ");
	Serial.println(Hall_R_02, 4);
	Serial.print("Hall_Z_02 ");
	Serial.println(Hall_Z_02, 4);
#endif // DEBUG

}
/*
	Calculate from bottom left of Drawing Area.

	First Arm pivot cooardinates:
		X = 100
		Y = 345

		X = 100 + Hall_R_01 + Hall_R_02
		Y = 345 - Hall_R_01 - Hall_R_02

*/
void Calculate_Coordinates() {

	Sylus_R = Hall_R_01 + Hall_R_02;
	Sylus_Z = 300.0 + Hall_Z_01 + Hall_Z_02;
	Sylus_A = Hall_Angle_03;

	lcd.setCursor(2, 0);
	lcd.print("  ");
	lcd.setCursor(Xpos(2, Sylus_R), 0);
	lcd.print(Sylus_R, 2);

	lcd.setCursor(11, 0);
	lcd.print("  ");
	lcd.setCursor(Xpos(11, Sylus_A), 0);
	lcd.print(Sylus_A, 1);

	lcd.setCursor(2, 1);
	lcd.print("  ");
	lcd.setCursor(Xpos(2, Sylus_Z), 1);
	lcd.print(Sylus_Z, 2);

#ifdef DEBUG
	Serial.print("Sylus_R ");
	Serial.println(Sylus_R, 4);
	Serial.print("Sylus_A ");
	Serial.println(Sylus_A, 4);
	Serial.print("Sylus_Z ");
	Serial.println(Sylus_Z, 4);
#endif // DEBUG

}
/*
	This a function to calculate the position of value displayed on LCD.

		Check to see if value is hundreds, tens or single.
		To keep numbers alighned.
*/
byte Xpos(byte pos, byte number) {
	byte val = pos;
	if (number < 100) { val = pos + 1; }
	if (number < 10) { val = pos + 2; }
	if (number < 0) { val = pos - 1; }
	return val;
}
/*
	When Pen Button is pressed.
*/
void Do_Button() {

	Button = false;

	//delay(200);	//	Uncomment if you don't have de-Bounce on the Sylus Switch
	Pen_Down = !Pen_Down;

	lcd.setCursor(12, 1);
	if (Pen_Down) { lcd.print("DOWN"); }
	else { lcd.print("UP  "); }

#ifdef DEBUG
	Serial.print("Pen_Down ");
	Serial.println(Pen_Down);
#endif // DEBUG

}
/*
	Send data to Serial Port.
	Want to send it fast, so it will be sent as raw bytes.

	Code is for ATMEGA based boards, where the double is only 4 bytes.

	Multiply by 1000 to get 3 decimal places as whole numbers.
	Max Dimention is 300mm
	so 300000 as binary = 00000100 10010011 11100000
	Needs 3 bytes for each dimention, total 6.
	Need another byte for four bits, -R, -Z, -A and pen state.

*/
void Send_Data() {

	byte _switch = 0;
	if (Pen_Down) { bitSet(_switch, 0); }
	if (Sylus_R < 0) { bitSet(_switch, 1); }
	if (Sylus_Z < 0) { bitSet(_switch, 2); }
	if (Sylus_A < 0) { bitSet(_switch, 3); }

	unsigned long _sylus_X = abs(Sylus_R * 1000);
	unsigned long _sylus_Y = abs(Sylus_Z * 1000);
	unsigned long _sylus_A = abs(Sylus_A * 1000);

	byte _buff[10];
	_buff[0] = _switch;					//	0 0 0 0 0 -X -Y PenDown
	_buff[1] = (byte)(_sylus_X >> 0);	//	X LSB
	_buff[2] = (byte)(_sylus_X >> 8);	//	
	_buff[3] = (byte)(_sylus_X >> 16);	//	X MSB
	_buff[4] = (byte)(_sylus_Y >> 0);	//	Y LSB
	_buff[5] = (byte)(_sylus_Y >> 8);	//	
	_buff[6] = (byte)(_sylus_Y >> 16);	//	Y MSM
	_buff[7] = (byte)(_sylus_A >> 0);	//	Y LSB
	_buff[8] = (byte)(_sylus_A >> 8);	//	
	_buff[9] = (byte)(_sylus_A >> 16);	//	Y MSM

#ifdef DEBUG
	for (size_t i = 0; i < 10; i++) {

		Serial.print("_buff ");
		Serial.println(_buff[i]);

	}
#else
	Serial.write(_buff, 10);
	Serial.println();
#endif // DEBUG

}
